"""Functions to be used by CI pipelines"""

import os
import shutil
from datetime import datetime
from os.path import basename, dirname, exists, isdir, isfile, join, splitext
from pathlib import Path
from typing import Optional, Dict

import dateutil.parser
import git

import urnc
from urnc.logger import critical, log, setup_logger, warn
from urnc.util import get_config_string, get_config_value, update_repo_config


def clone_student_repo(config: dict) -> git.Repo:
    """
    Clones the student repository if it doesn't exist locally, or returns the existing local repository.
    If the 'student' key is not found in the 'git' section of the config, it initializes a new student repository.

    Args:
        config (dict): The configuration dictionary.

    Returns:
        git.Repo: The cloned, existing, or newly initialized local git repository.

    Raises:
        Exception: If the 'student' key is not found in the 'git' section of the config and the initialization of a new repository fails.
        Exception: If the local repository exists but is not a git repository.
        Exception: If there is a mismatch between the remote URL and the local repository's URL.
    """

    # Get info about main repo
    main_path = urnc.util.get_course_root()
    main_name = basename(main_path)
    stud_url = get_config_string(config, "git", "student")

    # Init and return new repo if no student repo is specified
    if stud_url is None:
        stud_name = f"{main_name}-student"
        stud_path = main_path.parent.joinpath(stud_name)
        log(
            f"Property git.student not found in config. Initializing new student repo at {stud_path}"
        )
        stud_repo = git.Repo.init(stud_path)
        return stud_repo

    # Collect info about student repo if specified in config.yaml
    stud_name = splitext(basename(stud_url))[0]
    if stud_name == main_name:
        stud_name = f"{main_name}-student"
    stud_path = main_path.parent.joinpath(stud_name)

    # Return existing repo if already available at local filesystem
    if stud_path.exists():
        log(f"Returning existing repo {stud_path}")
        try:
            stud_repo = git.Repo(stud_path)
        except Exception:
            critical(f"Folder '{stud_path}' exists but is not a git repo")
        if stud_repo.remote().url != stud_url:
            critical(
                f"Repo remote mismatch. Expected: {stud_url}. Observed: {stud_repo.remote().url}."
            )
        stud_repo.remote().pull()
        return stud_repo

    # Clone and return repo if not available locally
    log(f"Cloning student repo {stud_url} to {stud_path}")
    stud_repo = git.Repo.clone_from(url=stud_url, to_path=stud_path)
    update_repo_config(stud_repo)
    return stud_repo


def clear_repo(repo):
    path = repo.working_dir
    entries = os.listdir(path)
    for entry in entries:
        if entry.startswith("."):
            continue
        entry_path = join(path, entry)
        if isfile(entry_path):
            os.remove(entry_path)
        if isdir(entry_path):
            shutil.rmtree(entry_path)


def copy_files(path, target_path):
    entries = os.listdir(path)
    for entry in entries:
        if entry.startswith("."):
            continue
        entry_path = join(path, entry)
        copy_path = join(target_path, entry)
        if isfile(entry_path):
            shutil.copy2(entry_path, copy_path)
        if isdir(entry_path):
            shutil.copytree(entry_path, copy_path)


def write_gitignore(
    main_gitignore: Optional[str], student_gitignore: str, config: Dict
) -> None:
    """
    Writes a ``.gitignore`` file in the student repository.

    This function copies the ``.gitignore`` file from the main repository (if it exists) and appends additional patterns to exclude based on the configuration.

    Parameters:
        main_gitignore (Optional[str]): The path to the .gitignore file in the main repository. If None, no .gitignore file is copied.
        student_gitignore (str): The path to the .gitignore file in the student repository.
        config (Dict): The configuration dictionary. Can contain a ``git`` key with a ``exclude`` key, which should be a list of patterns to exclude. Each pattern can be a string or a dictionary with a ``pattern`` key and optional ``after`` and ``until`` keys specifying a date range, e.g.::

                git:
                    exclude:
                        - *.pyc
                        - /tutorials/*.*
                        - {pattern: '!tutorials/Tutorial_1.ipynb', after: '2023-10-25 9:30 CET'}
                        - {pattern: '!tutorials/Tutorial21.ipynb', after: '2023-10-25 9:30 CET'}
    """
    if main_gitignore and exists(main_gitignore):
        shutil.copy(main_gitignore, student_gitignore)
    exclude = get_config_value(config, "git", "exclude", default=[])
    if not isinstance(exclude, list):
        critical("config.git.exclude must be a list")
    now = datetime.now()
    with open(student_gitignore, "a", newline="\n") as gitignore:
        gitignore.write("\n")
        for value in exclude:
            if isinstance(value, str):
                gitignore.write(f"{value}\n")
                continue
            if "after" in value and now < dateutil.parser.parse(value["after"]):
                continue
            if "until" in value and now > dateutil.parser.parse(value["until"]):
                continue
            gitignore.write(f"{value['pattern']}\n")


def update_index(repo):
    cached_files_str = repo.git.ls_files("-ci", "--exclude-standard")
    if cached_files_str != "":
        cached_files = cached_files_str.split("\n")
        log(f"Removing excluded files {cached_files}")
        repo.index.remove(cached_files, working_tree=False)
        repo.index.write()


def ci(commit=True):
    """
    Perform a continuous integration run on a student repository.

    This function

    1. Clones the student repository
    2. Clears it
    3. Copies over all files from the course repository
    4. Converts the notebooks
    5. Updates the .gitignore file according to the `git.exclude` setting from the `config.yaml`
    6. Commits and pushes the changes if `commit` is True

    Parameters:
        commit (bool): commit and push updates?

    Raises:
        Exception: If the repository is dirty and commit is True, an exception is raised.

    Examples:
        >>> ci(commit=True)   # This will commit and push the changes
        >>> ci(commit=False)  # This will not commit or push the changes
    """
    setup_logger(use_file=False)
    course_root = urnc.util.get_course_root()
    course_config = urnc.util.read_config()

    # Make sure main repo is clean if we want to commit and push to student remote
    if commit:
        course_repo = urnc.util.get_course_repo()
        if course_repo.is_dirty():
            raise Exception("Repo is not clean. Commit your changes first.")

    # Clone and clear student repo. Then copy over files from main repo
    student_repo = clone_student_repo(course_config)
    student_root = Path(student_repo.working_dir)
    clear_repo(student_repo)
    copy_files(course_root, student_root)

    # Convert notebooks
    solution_relpath = get_config_string(course_config, "ci", "solution", default=None)
    solution_pattern = (
        student_root.joinpath(solution_relpath) if solution_relpath else None
    )
    with urnc.util.chdir(student_root):
        # Paths printed as info messages by urnc.convert.convert are relative to the current working directory, so we change the working directory to the student repo root in order to get the shorter paths in the log messages.
        urnc.convert.convert(
            input=student_root,
            output=student_root,
            solution=solution_pattern,
            force=True,
            dry_run=False,
        )
    log("Notebooks converted")

    # Update .gitignore and drop cached files
    log("Updating .gitignore from config")
    write_gitignore(
        main_gitignore=course_root.joinpath(".gitignore"),
        student_gitignore=student_root.joinpath(".gitignore"),
        config=course_config,
    )
    log("Dropping cached files...")
    update_index(student_repo)

    # Commit and push
    if commit:
        log("Adding files and commiting")
        student_repo.git.add(all=True)
        student_repo.index.commit("urnc convert")
        log("Pushing student repo")
        student_repo.git.push()
        log("Done.")
    else:
        log("Skipping git commit and push")
        log("Done.")



"""Main entry point for urnc when called from command line"""

#!/usr/bin/env python3
import os
import click
import urnc


@click.group(help="Uni Regensburg Notebook Converter")
@click.option(
    "-f",
    "--root",
    help="Root folder for resolving relative paths. E.g. `urnc -f some/long/path convert xyz.ipynb out` is the same as `urnc convert some/long/path/xyz.ipynb some/long/path/out`.",
    default=os.getcwd(),
    type=click.Path(),
)
@click.version_option(prog_name="urnc", message="%(version)s")
@click.pass_context
def main(ctx, root):
    ctx.ensure_object(dict)
    ctx.obj["root"] = root


@click.command(
    short_help="Build student version and push to public repo",
    help="Run the urnc ci pipeline, i.e. create student versions of all notebooks and push the converted notebooks to the public repo. To test the pipeline locally, without actually pushing to the remote, use command `urnc student`. For further details see https://spang-lab.github.io/urnc/urnc.html#urnc.ci.ci.",
)
@click.pass_context
def ci(ctx):
    with urnc.util.chdir(ctx.obj["root"]):
        urnc.ci.ci(True)


@click.command(
    name="convert",
    short_help="Convert notebooks",
    help="Convert notebooks to student version. For details see https://spang-lab.github.io/urnc/urnc.html#urnc.convert.convert.",
)
@click.argument("input", type=click.Path(exists=True), default=".")
@click.argument("output", type=str, default="out")
@click.option("-s", "--solution", type=str, default=None)
@click.option("-v", "--verbose", is_flag=True)
@click.option("-f", "--force", is_flag=True)
@click.option("-n", "--dry-run", is_flag=True)
@click.pass_context
def convert(ctx, input, output, solution, verbose, force, dry_run):
    with urnc.util.chdir(ctx.obj["root"]):
        urnc.logger.setup_logger(False, verbose)
        urnc.convert.convert(input, output, solution, force, dry_run)


@click.command(help="Check notebooks for errors")
@click.argument("input", type=click.Path(exists=True), default=".")
@click.pass_context
@click.option("-q", "--quiet", is_flag=True)
def check(ctx, input, quiet):
    with urnc.util.chdir(ctx.obj["root"]):
        urnc.logger.setup_logger(use_file=False, verbose=not quiet)
        urnc.convert.convert(
            input=input,
            output="out",
            solution=None,
            force=False,
            dry_run=True,
            ask=False,
        )


@click.command(help="Manage the semantic version of your course")
@click.option("--self", is_flag=True, help="Echo the version of urnc")
@click.argument(
    "action",
    type=click.Choice(["show", "patch", "minor", "major"]),
    required=False,
    default="show",
)
@click.pass_context
def version(ctx, self, action):
    with urnc.util.chdir(ctx.obj["root"]):
        urnc.logger.setup_logger(use_file=False, verbose=False)
        if self:
            urnc.version.version_self(action)
        else:
            urnc.version.version_course(action)


@click.command(help="Tag the current commit with the current version of the course")
@click.pass_context
def tag(ctx):
    with urnc.util.chdir(ctx.obj["root"]):
        urnc.version.tag()


@click.command(help="Build and show the student version")
@click.pass_context
def student(ctx):
    with urnc.util.chdir(ctx.obj["root"]):
        urnc.ci.ci(commit=False)


@click.command(help="Pull the course repo")
@click.argument("course_name", type=str, default=None, required=False)
@click.option(
    "-o", "--output", type=str, help="The name of the output folder", default=None
)
@click.option("-b", "--branch", help="The branch to pull", default="main")
@click.option("-d", "--depth", help="The depth for git fetch", default=1)
@click.pass_context
def pull(ctx, course_name, output, branch, depth):
    with urnc.util.chdir(ctx.obj["root"]):
        urnc.logger.setup_logger()
        try:
            urnc.pull.pull(course_name, output, branch, depth)
        except Exception as err:
            urnc.logger.error("pull failed with unexpected error.")
            urnc.logger.error(err)


main.add_command(version)
main.add_command(convert)
main.add_command(ci)
main.add_command(check)
main.add_command(student)
main.add_command(pull)
main.add_command(tag)



"""Create student version of one or more Jupyter notebooks"""

import os
import sys
from pathlib import Path
from typing import List, Optional, Union

import nbformat
import traitlets
from nbconvert.exporters.notebook import NotebookExporter
from nbconvert.preprocessors.clearoutput import ClearOutputPreprocessor

import urnc.logger as log
from urnc.preprocessor.add_tags import AddTags
from urnc.preprocessor.lint import Linter
from urnc.preprocessor.solutions import SolutionRemover, SkeletonRemover


student_preprocessors = [Linter, AddTags, SolutionRemover, ClearOutputPreprocessor]
student_config = traitlets.config.Config()
student_config.NotebookExporter.preprocessors = student_preprocessors
student_converter = NotebookExporter(student_config)

solution_preprocessors = [AddTags, SkeletonRemover, ClearOutputPreprocessor]
solution_config = traitlets.config.Config()
solution_config.NotebookExporter.preprocessors = solution_preprocessors
solution_converter = NotebookExporter(solution_config)


class NbPath(object):
    def __init__(self, path: Union[str, Path], rootpath: Union[str, Path]):
        """Path to a jupyter notebook inside a arbitrarily nested directory structure.

        Attributes:
            path (Path): Path object initialized from the provided path.
            abspath (str): Absolute path to the notebook.
            absdirpath (str): Absolute path to the directory containing the notebook.
            rootpath (str): Absolute path to the root directory of the notebook.
            relpath (str): Relative path to the notebook from the root directory.
            reldirpath (str): Relative path to the directory containing the notebook from the root directory.
            name (str): Name of the notebook including extension.
            basename (str): Name of the notebook without extension.
            ext (str): Extension of the notebook without the dot.

        Example:
            >>> path = "C:/Users/max/mycourse/lectures/week2/lecture1.ipynb"
            >>> rootpath = "C:/Users/max/mycourse"
            >>> nb = NbPath(path, rootpath)
            >>> nb.abspath    == "C:/Users/max/mycourse/lectures/week2/lecture1.ipynb"
            >>> nb.absdirpath == "C:/Users/max/mycourse/lectures/week2"
            >>> nb.rootpath   == "C:/Users/max/mycourse"
            >>> nb.relpath    == "lectures/week2/lecture1.ipynb"
            >>> nb.reldirpath == "lectures/week2"
            >>> nb.name       == "lecture1.ipynb"
            >>> nb.basename   == "lecture1"
            >>> nb.ext        == "ipynb"
        """
        self.path = path = Path(path)
        self.rootpath = rootpath = Path(rootpath)
        assert path.exists(), f"Notebook {path} does not exist"
        assert rootpath.exists(), f"Directory {rootpath} does not exist"
        self.abspath = path.absolute()
        self.relpath = path.relative_to(rootpath)
        self.absdirpath = path.parent.absolute()
        self.reldirpath = self.absdirpath.relative_to(self.rootpath)
        self.name = path.name
        self.basename = path.stem
        self.ext = path.suffix[1:] if path.suffix.startswith(".") else path.suffix


def convert(
    input: str = ".",
    output: str = "out/",
    solution: Optional[str] = None,
    force: bool = False,
    dry_run: bool = False,
    ask: bool = True,
) -> None:
    """Convert one or multiple notebooks to its student version with and/or without solutions removed.

    Args:
        input: The input path. Can be either the path to a single ipynb file or the path to a directory containing ipynb files.
        output: Path for storing the student notebooks with solutions removed. Directories will be created if necessary. If ``output`` is None, the conversion will be skipped. Supports placeholder variables. For details see section "Details" below.
        solution: Path for storing the student notebooks with solutions kept. Directories will be created if necessary. If ``solution`` is None, the conversion will be skipped. Supports placeholder variables. For details see section "Details" below.
        force: If true, existing output files will be overwritten.
        dry_run: If true, input files will be converted but not written to disk. This is useful for checking if the conversion works without changing any files.
        ask: Should the user be asked if existing files should be overwritten?

    Details:
        A call like ``convert("C:/Users/max/mycourse", ...)`` will search ``C:/Users/max/mycourse`` recursively for notebooks and convert each notebook to student versions. To be able to specify output paths in a simple yet flexible way, the following placeholder variables are provided: ``nb.abspath``, ``nb.absdirpath``, ``nb.rootpath``, ``nb.relpath``, ``nb.reldirpath``, ``nb.name``, ``nb.basename`` and ``nb.ext``. To understand the meaning of each placeholder mean, consider the following folder structure::

            C:/Users/max/mycourse
            ├── lectures
            │   ├── week1
            │   │   ├── lecture1.ipynb
            │   │   └── lecture2.ipynb
            │   └── week2
            │       ├── lecture3.ipynb
            │       └── lecture4.ipynb
            └── assignments
                ├── week1.ipynb
                └── week2.ipynb

        The above course contains the following six notebooks: ``lecture1.ipynb``, ``lecture2.ipynb``, ``lecture3.ipynb``, ``lecture4.ipynb``, ``week1.ipynb`` and ``week2.ipynb``. The following table shows the value of each placeholder variable exemplarily for notebook ``C:/Users/max/mycourse/lectures/week2/lecture1.ipynb``::

            | nb.abspath    | C:/Users/max/mycourse/lectures/week2/lecture1.ipynb |
            | nb.absdirpath | C:/Users/max/mycourse/lectures/week2                |
            | nb.rootpath   | C:/Users/max/mycourse                               |
            | nb.relpath    |                       lectures/week2/lecture1.ipynb |
            | nb.reldirpath |                       lectures/week2/               |
            | nb.name       |                                      lecture1.ipynb |
            | nb.basename   |                                      lecture1       |
            | nb.ext        |                                               ipynb |

        I.e. the following function call:

            >>> convert(
            >>>     input = "C:/Users/max/mycourse",
            >>>     output = "C:/converted/{nb.relpath}",
            >>>     solution = "C:/converted/{nb.reldirpath}/solutions/{nb.name}",
            >>> )

        would create the following output structure::

            C:/converted
            ├── lectures
            │   ├── week1
            │   │   ├── lecture1.ipynb
            │   │   ├── lecture1.ipynb
            │   │   └── solutions
            │   │       └── lecture1.ipynb
            │   │       └── lecture2.ipynb
            │   └── week2
            │       ├── lecture3.ipynb
            │       ├── lecture4.ipynb
            │       └── solutions
            │           └── lecture3.ipynb
            │           └── lecture4.ipynb
            └── assignments
                ├── week1.ipynb
                ├── week2.ipynb
                └── solutions
                    ├── week1.ipynb
                    └── week2.ipynb

        To make specifying output paths less verbose and to maintain backwards compatibility, the following shortcuts are provided:

        1. If ``output`` does not end with ``.ipynb`` and does not contain any placeholders, the provided path will be interpreted as a directory and the notebooks will be saved in ``{output}/{nb.relpath}``, i.e. the following calls are equivalent::

            convert('mycourse', output = 'out')
            convert('mycourse', output = 'out/{nb.relpath}')

        2. If ``solution`` does not end with ``.ipynb`` and does not contain any placeholders, the provided path will be interpreted as a directory and the notebooks will be saved in ``{solution}/{nb.reldirpath}/{nb.basename}-solution.{nb.ext}``, i.e. the following calls are equivalent::

                convert('mycourse', solution = 'out')
                convert('mycourse', solution = 'out/{nb.reldirpath}/{nb.basename}-solution.{nb.ext}')

    Examples:
        >>> convert("/path/to/single_notebook.ipynb", "/path/to/output", dry_run=True)
        >>> convert("/path/to/notebooks", "/path/to/output", force=True)
    """
    log.log(f"Converting notebooks in {os.path.abspath(input)}")
    nbs = get_nb_paths(input)
    outstr = str(output).lower() if output else "{}"
    solstr = str(solution).lower() if solution else "{}"
    if not ("{" in outstr or outstr.endswith(".ipynb")):
        output = f"{output}/{{nb.relpath}}"
    if not ("{" in solstr or solstr.endswith(".ipynb")):
        solution = f"{solution}/{{nb.reldirpath}}/{{nb.basename}}-solution.{{nb.ext}}"
    for nb in nbs:
        inb = nb.abspath
        onb = Path(str(output).format(nb=nb)) if output else None
        snb = Path(str(solution).format(nb=nb)) if solution else None
        convert_nb(
            input=inb, output=onb, solution=snb, force=force, dry_run=dry_run, ask=ask
        )


def convert_nb(
    input: Union[str, Path],
    output: Optional[Union[str, Path]] = None,
    solution: Optional[Union[str, Path]] = None,
    force: bool = False,
    dry_run: bool = False,
    ask: bool = True,
) -> None:
    """Convert a single notebook to its student version with and/or without solutions removed.

    Args:
        input: Path of input notebook.
        output: Path for storing the student notebook without solutions. If the parent directory does not exist, it will be created. If output is None, the conversion will be skipped.
        solution: Path for storing the student notebook with solutions. If the parent directory does not exist, it will be created. If solution is None, the conversion will be skipped.
        force: Overwrite existing output files without asking?
        dry_run: Skip writing to disk after performing conversions? Useful for testing without file changes.
        ask: Should the user be asked if existing files should be overwritten?

    Examples:
        >>> # Create student version with solutions removed/kept from dummy.ipynb
        >>> convert_nb("dummy.ipynb", output="dummy-converted.ipynb")
        >>> convert_nb("dummy.ipynb", solution="dummy-solution.ipynb")
        >>>
        >>> # Same as above but in one line
        >>> convert_nb("dummy.ipynb", output="dummy-converted.ipynb", solution="dummy-solution.ipynb")
        >>>
        >>> # Absolute paths work as well
        >>> convert_nb("C:/Users/max/dummy.ipynb", solution="C:/Users/max/dummy-solution.ipynb")
        >>>
        >>> # Overwrite existing files without asking
        >>> convert_nb("dummy.ipynb", output="dummy-converted.ipynb", force=True)
        >>>
        >>> # Don't write converted files to disk
        >>> convert_nb("dummy.ipynb", output="dummy-converted.ipynb", dry_run=True)
    """
    cwd = os.getcwd()
    in_path = Path(input)
    in_relpath = in_path.relative_to(cwd) if in_path.is_relative_to(cwd) else in_path
    out_path = Path(output) if output else None
    sol_path = Path(solution) if solution else None
    log.log(f"Reading {in_relpath}")
    in_text = nbformat.read(in_path, as_version=4)
    resources = {"path": in_path}
    if out_path:
        log.log(f"Converting {in_relpath} to student version without solutions")
        out_text = student_converter.from_notebook_node(in_text, resources)[0]
        if not dry_run:
            write(text=out_text, path=out_path, force=force, ask=ask)
    if sol_path:
        log.log(f"Converting {in_relpath} to student version with solutions")
        sol_text = solution_converter.from_notebook_node(in_text, resources)[0]
        if not dry_run:
            write(text=sol_text, path=sol_path, force=force, ask=ask)


def get_nb_paths(input: Union[str, Path]) -> List[NbPath]:
    """Creates `NbPath` objects from all notebook files in input and returns them as list."""
    input = Path(os.path.abspath(input))
    nbs = []
    if input.is_file():
        nb = NbPath(path=input, rootpath=input.parent)
        nbs.append(nb)
    elif input.is_dir():
        for root, dirs, files in os.walk(input, topdown=True):
            dirs[:] = [d for d in dirs if not d[0] == "."]
            files[:] = [f for f in files if f.lower().endswith(".ipynb")]
            for file in files:
                nb = NbPath(path=os.path.join(root, file), rootpath=input)
                nbs.append(nb)
    else:
        log.warn(f"Input path {input} does not exist")
    return nbs


def write(text: str, path: Union[str, Path], force: bool = False, ask=True) -> None:
    """
    Write text to path.
    If the parent directory does not exist, it will be created.
    If the path exists already and force is False, the user will be asked if they want to overwrite the file.
    If stdout is not a tty or ask is False, a warning message will be printed instead.
    If force is True, the file will be overwritten without asking.

    Args:
        text: The text to write.
        path: The path of the file to write to. If the parent directory does not exist, it will be created.
        force: Overwrite existing files without asking?
        ask: Ask the user if they want to overwrite the file if it exists already and force is False?
    """
    cwd = os.getcwd()
    path = Path(path)
    relpath = path.relative_to(cwd) if path.is_relative_to(cwd) else path
    path.parent.mkdir(parents=True, exist_ok=True)
    if not path.exists() or force:
        log.log(f"Writing {relpath}.")
        with open(path, "w", newline="\n") as f:
            f.write(text)
        return
    if not sys.stdout.isatty() or not ask:
        log.log(
            f"Skipped writing {relpath} because it exists already and force is False."
        )
        return
    overwrite = ""
    while overwrite.lower() not in ["y", "n"]:
        overwrite = input(f"File {path} exists. Overwrite? (y/n): ")
    if overwrite.lower() == "y":
        with open(path, "w", newline="\n") as f:
            f.write(text)



"""
University of Regensburg Notebook Converter

Package for converting jupyter notebook files within a course directory into a
format suitable for student use. It simplifies the process of preparing and
maintaining course materials by providing easy-to-use commands to check,
convert, and manage course versions.
"""

from urnc import ci, convert, logger, pull, util, version



"""Manage the semantic version of your course"""

import click

import semver
import urnc.logger as log

import urnc.util as util


def bump(version: str, action: str):
    """Bump the version based on the action

    :param version: The current version.
    :param action: The action to perform (show, patch, minor, major, prerelease).
    :return: The new version after bumping, or None if action is 'show'.
    """
    v = semver.Version.parse(version)
    print(v)
    match action:
        case "show":
            return
        case "patch":
            return v.bump_patch()
        case "minor":
            return v.bump_minor()
        case "major":
            return v.bump_major()
        case "prerelease":
            return v.bump_prerelease()
    raise click.UsageError("Invalid action")


def version_self(action: str) -> None:
    """Bump the version of the urnc project itself.

    :param action: The action to perform (show, patch, minor, major).
    """
    repo = util.get_urnc_repo()
    config = util.read_pyproject()
    v = config["project"]["version"]
    new_version = bump(v, action)
    if not new_version:
        return
    if repo is not None and repo.is_dirty():
        log.error("Repo is not clean, commit your changes before calling version")
        return
    log.log(f"New Version: {new_version}")
    config["project"]["version"] = str(new_version)
    path = util.write_pyproject(config)
    if repo is not None:
        message = f"v{new_version}"
        log.log("Committing new version and tagging the commit...")
        repo.index.add(path)
        repo.index.commit(message)
        repo.create_tag(message, "HEAD", message)
        print("Done.")


def version_course(action: str) -> None:
    """Bump the version of the course.

    :param action: The action to perform (show, patch, minor, major).
    """
    try:
        config = util.read_config()
    except Exception:
        raise click.UsageError(
            "Could not find a config.yaml in the current directory. Use 'urnc version --self' to get the version of urnc."
        )
    v = config["version"]
    new_version = bump(v, action)
    if new_version:
        print(f"New Version: {new_version}")
        config["version"] = str(new_version)
        util.write_config(config)


def tag():
    """Tag the current commit with the current version of the course."""
    repo = util.get_course_repo()
    config = util.read_config()
    v = config["version"]
    tag = f"v{v}"
    if util.tag_exists(repo, tag):
        raise click.UsageError(
            f"Tag {tag} exists already. Make sure to increment the version in config.yaml."
        )
    repo.create_tag(tag, "HEAD", tag)
    log.log(f"Tagged the current commit with {tag}.")



from typing import Tuple, Dict, Union
from nbconvert.preprocessors.base import Preprocessor

import urnc.preprocessor.util as util
from urnc.preprocessor.util import Tags, Keywords
import re

from nbformat.notebooknode import NotebookNode


def remove_solution_lines(cell: NotebookNode) -> Union[NotebookNode, None]:
    """
    Remove solution lines from a cell.

    This function takes a cell from a Jupyter Notebook as input, removes solution lines, uncomments skeleton lines and returns the cell.
    If no lines remain after removing the solutions, the function returns None (i.e. the whole cell is removed).
    If no Solution or Skeleton patterns are found in the cell, the function returns None as well.
    Solutions and Skeleton pattern are defined at `urnc.preprocessor.util.Keywords <https://spang-lab.github.io/urnc/urnc.preprocessor.html#urnc.preprocessor.util.Keywords>`_.

    Parameters:
        cell: The cell to process.

    Returns:
        Either the cell without solutions lines and uncommented skeleton lines or None.
    """
    text = cell.source
    lines = text.split("\n")
    last_tag = None
    found_tag = False
    tagged_lines = []
    for line in lines:
        if re.match(Keywords.SOLUTION, line, re.IGNORECASE):
            last_tag = Tags.SOLUTION
            found_tag = True
            continue
        if re.match(Keywords.SKELETON, line, re.IGNORECASE):
            last_tag = Tags.SKELETON
            found_tag = True
            continue
        if re.match(Keywords.SOLUTION_END, line, re.IGNORECASE):
            last_tag = None
            continue
        tagged_lines.append((line, last_tag))
    if not found_tag:
        return None
    skeleton_lines = []
    for line, tag in tagged_lines:
        if tag is None:
            skeleton_lines.append(line)
        if tag == Tags.SOLUTION:
            continue
        if tag == Tags.SKELETON:
            uncom = re.sub(r"^#\s?", "", line)
            skeleton_lines.append(uncom)
    if len(skeleton_lines) == 0:
        return None
    cell.source = "\n".join(skeleton_lines)
    return cell


def remove_skeleton_lines(text: str) -> str:
    """
    Remove skeleton lines from a string.

    Parameters:
        text: The text to process.

    Returns:
        Text without skeleton lines.

    Examples::
        remove_skeleton_lines("### Skeleton\n# # a == ?") == ""
        remove_skeleton_lines("### Skeleton\n# # a == ?\n###") == "###"
        remove_skeleton_lines("### Solution\na = 100\n### Skeleton\n# # a == ?\n###") == "### Solution\na = 100\n###"
    """
    lines = text.split("\n")
    keep = []
    inside_skeleton = False
    for line in lines:
        if re.match(Keywords.SKELETON, line, re.IGNORECASE):
            inside_skeleton = True
        elif re.match(Keywords.SOLUTION_END, line, re.IGNORECASE):
            inside_skeleton = False
            keep.append(line)
        elif not inside_skeleton:
            keep.append(line)
    return "\n".join(keep)


class SolutionRemover(Preprocessor):
    def preprocess(self,
                   notebook: NotebookNode,
                   resources: Union[Dict, None] = None) -> Tuple[NotebookNode, Union[NotebookNode, None]]:
        """
        Preprocess the notebook by removing solutions lines from solution cells and uncommenting skeleton lines within solution cells.

        This function takes a notebook and resources as input, removes the solution lines from notebook cells  and returns the notebook and resources.

        Parameters:
            notebook: The notebook to preprocess.
            resources: The resources associated with the notebook.

        Returns:
            The preprocessed notebook and the resources.
        """
        cells = []
        for cell in notebook.cells:
            if util.has_tag(cell, Tags.SOLUTION):
                cell = remove_solution_lines(cell)
            if cell is not None:
                cells.append(cell)
        notebook.cells = cells
        return notebook, resources


class SkeletonRemover(Preprocessor):
    def preprocess(self,
                   notebook: NotebookNode,
                   resources: Union[Dict, None] = None) -> Tuple[NotebookNode, Union[NotebookNode, None]]:
        """
        Preprocess the notebook by removing skeleton lines from solution cells.

        This function takes a notebook and resources as input, removes the solution lines from notebook cells  and returns the notebook and resources.

        Parameters:
            notebook: The notebook to preprocess.
            resources: The resources associated with the notebook.

        Returns:
            The preprocessed notebook and the resources.
        """
        for cell in notebook.cells:
            if util.has_tag(cell, Tags.SOLUTION):
                cell.source = remove_skeleton_lines(text = cell.source)
        return notebook, resources


"""Fetch and merge the student repo"""

import os
import git
import datetime

import urnc.util as util

import urnc.logger as log


def get_upstream_changes(repo):
    branch_name = repo.active_branch
    changes = repo.git.diff(f"..origin/{branch_name}", "--name-status").split("\n")
    return changes


def get_upstream_deleted(repo):
    changes = get_upstream_changes(repo)
    files = []
    for change in changes:
        parts = change.split("\t", 1)
        if len(parts) != 2:
            continue
        [ctype, file] = change.split("\t", 1)
        if ctype == "D":
            files.append(file)
    return files


def get_upstream_added(repo):
    changes = get_upstream_changes(repo)
    files = []
    for change in changes:
        parts = change.split("\t", 1)
        if len(parts) != 2:
            continue
        [ctype, file] = parts
        if ctype == "A":
            files.append(file)
    return files


def get_course_info(course_name, output):
    base_url = os.getenv("GIT_URL")
    if not base_url:
        log.error("Failed to pull. Missing env var GIT_URL")
        return (None, None)

    if not course_name:
        try:
            git_repo = git.Repo(os.getcwd(), search_parent_directories=True)
            git_url = git_repo.remote().url
            if not git_url.startswith(base_url):
                log.error(f"Failed to pull. Invalid git remote_url {git_url}")
                return (None, None)
            folder_name = git_repo.working_dir
            return (git_url, folder_name)
        except Exception:
            log.error("Failed to pull no course_name given and not in a git repo")
            return (None, None)
    folder_name = output
    if not output:
        folder_name = course_name

    git_url = f"{base_url}{course_name}.git"
    return (git_url, folder_name)


def get_repo(folder_name):
    base_url = os.getenv("GIT_URL")
    if not base_url:
        log.error("Missing env var GIT_URL")
        return None
    try:
        git_repo = git.Repo(folder_name)
        git_url = git_repo.remote().url
        if not git_url.startswith(base_url):
            log.error(
                f"Invalid git remote_url {git_url} needs to start with {base_url}"
            )
            return None
        return git_repo
    except Exception:
        log.error(f"{folder_name} is not a git repo")
        return None


def reset_deleted_files(repo):
    branch = repo.active_branch
    deleted_files = repo.git.ls_files("--deleted").split("\n")
    deleted_upstream = get_upstream_deleted(repo)
    for filename in deleted_files:
        if not filename:
            continue
        if filename in deleted_upstream:
            repo.git.checkout("HEAD", "--", filename)
        else:
            log.dbg(f"Restoring deleted file {filename}")
            repo.git.checkout(f"origin/{branch}", "--", filename)


def rename_file_with_timestamp(path):
    timestamp = datetime.datetime.now().strftime("%Y%m%d%H%M%S")
    (folder, filename) = os.path.split(path)
    (basename, ext) = os.path.splitext(filename)
    new_filename = f"{basename}_{timestamp}{ext}"
    new_path = os.path.join(folder, new_filename)
    log.warn(
        f"Renaming {path} to {new_path}, to avoid merge conflict with local untracked files."
    )
    os.rename(path, new_path)


def rename_local_untracked(repo):
    added_files = get_upstream_added(repo)
    for filename in added_files:
        if not filename:
            continue
        path = os.path.join(repo.working_dir, filename)
        if os.path.exists(path):
            rename_file_with_timestamp(path)


def merge(repo):
    branch = repo.active_branch
    remote_branch = f"origin/{branch}"
    try:
        repo.git.merge("-Xours", remote_branch)
    except git.GitCommandError as err:
        if "CONFLICT (modify/delete)" in str(err):
            log.warn(
                "Found a CONFLICT (modify/delete). Keeping the local file by commiting."
            )
            repo.git.commit("-am", "Resolve CONFLICT (modify/delete)", "--allow-empty")
            return
        raise


def pull_admin(course_name, output, branch, depth):
    isAdmin = os.getenv("IS_ADMIN") == "1"
    if not isAdmin:
        return
    if not course_name:
        return
    base_url = os.getenv("GIT_URL_ADMIN")
    if not base_url:
        log.error("Missing env var GIT_URL_ADMIN")
        return
    git_url = f"{base_url}{course_name}.git"
    folder_name = output
    if not output:
        folder_name = f"{course_name}-admin"

    if not os.path.exists(folder_name):
        log.log(f"{folder_name} does not exists. Cloning repo {git_url}")
        try:
            git.Repo.clone_from(git_url, folder_name, branch=branch, depth=depth)
            log.log("Cloned successfully.")
        except Exception as err:
            log.error("Failed to clone repo. Error:")
            log.error(err)
    else:
        log.log(f"{folder_name} exists.")
        try:
            repo = git.Repo(folder_name)
            log.log("Pulling...")
            repo.git.pull("--ff-only")
            log.log("Done")

        except Exception:
            log.error(f"{folder_name} is not a git repo")


def pull(course_name, output, branch, depth):
    pull_admin(course_name, output, branch, depth)

    (git_url, folder_name) = get_course_info(course_name, output)
    if not git_url or not folder_name:
        return

    if not os.path.exists(folder_name):
        log.log(f"{folder_name} does not exists. Cloning repo {git_url}")
        try:
            git.Repo.clone_from(git_url, folder_name, branch=branch, depth=depth)
            log.log("Cloned successfully.")
        except Exception as err:
            log.error("Failed to clone repo. Error:")
            log.error(err)
        return
    log.log(f"{folder_name} exists. Updating the repo")
    repo = get_repo(folder_name)
    if not repo:
        return

    log.log("Fetching changes...")
    repo.remote().fetch()
    log.log("Checking for local untracked files")
    rename_local_untracked(repo)
    log.log("Restoring locally deleted files")
    reset_deleted_files(repo)
    log.log("Unstaging all changes")
    repo.git.reset("--mixed")

    util.update_repo_config(repo)
    if repo.is_dirty():
        log.log("Repo is dirty. Commiting....")
        repo.git.commit("-am", "Automatic commit by urnc", "--allow-empty")
        log.log("Created new commit")

    log.log("Merging from remote...")
    merge(repo)
    log.log("Done.")



"""Utility functions for urnc"""

import abc
import os
import re
from pathlib import Path
from typing import Optional

import click
import git
import tomli_w
from ruamel.yaml import YAML

try:
    import tomllib
except Exception:
    tomllib = None

yaml = YAML(typ="rt")


# Git related functions


def branch_exists(repo, branch):
    origin_branch = f"origin/{branch}"
    for ref in repo.references:
        if ref.name == branch or ref.name == origin_branch:
            return True
    return False


def tag_exists(repo, tag):
    ref = f"refs/tags/{tag}"
    try:
        repo.git.rev_parse("--quiet", "--verify", ref)
        return True
    except Exception:
        return False


def update_repo_config(repo):
    try:
        r = repo.config_reader()
        name = r.get_value("user", "name")
        email = r.get_value("user", "email")
        assert name is not None
        assert email is not None
        return
    except Exception:
        pass

    config_writer = repo.config_writer()
    config_writer.set_value("user", "name", "urnc")
    config_writer.set_value("user", "email", "urnc@spang-lab.de")
    config_writer.release()


def get_course_repo():
    path = get_course_root()
    try:
        git_repo = git.Repo(path, search_parent_directories=False)
        return git_repo
    except Exception:
        raise click.UsageError(f"Path '{path}' is not a git repo")


def get_urnc_repo():
    path = get_urnc_root()
    try:
        git_repo = git.Repo(path, search_parent_directories=False)
        return git_repo
    except Exception:
        raise click.UsageError(f"Path '{path}' is not a git repo")


# Functions for finding/switching course/urnc root paths


def get_urnc_root():
    """
    This function navigates up the directory tree from the current working directory until it finds a 'pyproject.toml' file that belongs to the 'urnc' package, or it reaches the root directory. If it doesn't find a suitable    'pyproject.toml' file, it raises an Exception.

    Returns:
        str: The path of the directory containing the 'pyproject.toml' file of the 'urnc' package.

    Raises:
        Exception: If no 'pyproject.toml' file is found in the directory hierarchy.
        Exception: If the found 'pyproject.toml' file does not belong to the 'urnc' package.
    """
    path = Path(os.getcwd())
    while path != path.parent:
        pyproject_toml = path.joinpath("pyproject.toml")
        if pyproject_toml.exists():
            pyproject = open(pyproject_toml).read()
            if re.search(r'name\s*=\s*"?urnc"?', pyproject):
                return path
            raise Exception(f"Path '{path}' does not belong to the 'urnc' package")
        path = path.parent
    raise Exception("No 'pyproject.toml' found in the directory hierarchy")


def get_course_root() -> Path:
    """
    This function navigates up the directory tree from the current working directory until it finds a `config.yaml` file or it reaches the root directory. If it doesn't find a 'config.yaml' file, it raises an Exception.

    Returns:
        Path: path of the course directory.

    Raises:
        Exception: If no 'config.yaml' file is found in the directory hierarchy.
    """
    path = Path(os.getcwd())
    while path != path.parent:
        if path.joinpath("config.yaml").exists():
            return path
        path = path.parent
    raise Exception("No 'config.yaml' found in the directory hierarchy")


class chdir(abc.ABC):
    """Non thread-safe context manager to change the current working directory.

    Same as `contextlib.chdir <https://docs.python.org/3/library/contextlib.html#contextlib.chdir>`_. We implement it ourselves because ``contextlib.chdir`` is only available for python versions >=3.11 and we want to support versions >=3.8.

    Example:
        >>> with chdir(".."):
        >>>     # This code will be executed in the parent directory of the current working directory
        >>>     pass
    """

    def __init__(self, path):
        self.path = path
        self._old_cwd = []

    def __enter__(self):
        self._old_cwd.append(os.getcwd())
        os.chdir(self.path)

    def __exit__(self, exc_type, exc_value, traceback):
        os.chdir(self._old_cwd.pop())


# Functions for reading/writing course configs


def read_config(course_root: Optional[Path] = None) -> dict:
    """
    Reads the configuration from a YAML file named 'config.yaml' located at the root of the git repository.

    Args:
        course_root: The root directory of the course. Defaults to the root of the git repository when called from within a git repository.

    Raises:
        click.UsageError: If the 'config.yaml' file does not exist in the git root folder.
        click.FileError: If there is an error reading the file or processing its content.

    Returns:
        dict: The configuration dictionary.
    """
    filename = "config.yaml"
    if course_root is None:
        course_root = get_course_root()
    path = os.path.join(course_root, filename)
    if not os.path.isfile(path):
        raise click.UsageError(
            f"urnc expects a config file called '{filename}' in the course root folder '{course_root}'"
        )
    try:
        with open(path, "r") as f:
            config = yaml.load(f)
        if "git" in config and "student" in config["git"]:
            config["git"]["student"] = config["git"]["student"].format(**os.environ)
    except Exception as e:
        raise click.FileError(path, str(e))
    return config


def write_config(data):
    filename = "config.yaml"
    base_path = get_course_root()

    path = os.path.join(base_path, filename)
    try:
        with open(path, "w", newline="\n") as f:
            yaml.dump(data, f)
            return path
    except Exception as e:
        raise click.FileError(path, str(e))


def get_config_value(config, *args, default=None, required=False):
    value = config
    full_key = "config"
    for key in args:
        full_key = f"{full_key}.{key}"
        if key not in value:
            if required:
                raise Exception(f"{full_key} is required")
            return default
        value = value[key]
    return value


def get_config_string(config, *args, default=None, required=False):
    value = get_config_value(config, *args, default=default, required=required)
    if value is None:
        return default
    if not isinstance(value, str):
        full_key = ".".join(args)
        raise Exception(f"config value config.{full_key} must be a string")
    return str(value)


# Functions for reading/writing urnc's pyproject.toml


def read_pyproject():
    if tomllib is None:
        raise click.UsageError("tomllib (python3.11) is required")
    filename = "pyproject.toml"
    base_path = get_urnc_root()
    path = os.path.join(base_path, filename)
    try:
        with open(path, "rb") as f:
            config = tomllib.load(f)
            return config
    except Exception as e:
        raise click.FileError(path, str(e))


def write_pyproject(data):
    filename = "pyproject.toml"
    base_path = get_urnc_root()
    path = os.path.join(base_path, filename)
    try:
        with open(path, "wb") as f:
            tomli_w.dump(data, f)
            return path
    except Exception as e:
        raise click.FileError(path, str(e))



from nbconvert.preprocessors.base import Preprocessor

import re
import urnc.preprocessor.util as util
from urnc.preprocessor.util import Keywords, Tags
import urnc.logger as log


def is_assignment_start(cell):
    if util.has_tag(cell, Tags.NORMAL):
        return None
    opts = re.IGNORECASE | re.MULTILINE
    if match := re.search(Keywords.ASSIGNMENT_DEPRECATED, cell.source, opts):
        id = util.to_snake_case(match.group(1))
        return id
    if match := re.search(Keywords.ASSIGNMENT_START, cell.source, opts):
        id = util.to_snake_case(match.group(1))
        return id
    return None


def is_solution(cell):
    if util.has_tag(cell, Tags.NORMAL):
        return False
    if util.has_tag(cell, Tags.SOLUTION):
        return True
    opts = re.IGNORECASE | re.MULTILINE
    return re.search(Keywords.SOLUTION, cell.source, opts)


def is_assignment_end(cell, assignment_id):
    if util.has_tag(cell, Tags.NORMAL):
        return False
    if assignment_id is None:
        return False
    if cell.cell_type != "markdown":
        return False
    if is_solution(cell):
        return False
    if re.search(Keywords.HEADER, cell.source, re.MULTILINE):
        return True
    return False


class AddTags(Preprocessor):
    def preprocess(self, nb, resources):
        assignment_ids = set()
        assignment_id = None
        has_solution = False

        for cell in nb.cells:
            if is_assignment_end(cell, assignment_id):
                if not has_solution:
                    log.warn(f"Assignment {assignment_id} has no solution")
                assignment_id = None

            if id := is_assignment_start(cell):
                assignment_id = id
                has_solution = False
                util.set_tag(cell, Tags.ASSIGNMENT_START)
                if assignment_id in assignment_ids:
                    log.warn("Duplicate Assignment id '%s'" % assignment_id)
                else:
                    assignment_ids.add(assignment_id)
                log.dbg("Detected Assignment '%s'" % assignment_id)

            if assignment_id:  # check if cell is part of the assignment
                util.set_tag(cell, Tags.ASSIGNMENT)
                cell.metadata.assignment_id = assignment_id

            if is_solution(cell):
                preview = util.cell_preview(cell)
                if assignment_id is None:
                    log.dbg(f"Solution cell is not part of an assignment: {preview}")
                log.dbg(f"Detected Solution cell {preview}")
                util.set_tag(cell, Tags.SOLUTION)
                has_solution = True

        if assignment_id is not None and not has_solution:
            log.warn(f"Assignment {assignment_id} has no solution")

        return nb, resources



import enum
import re


class Keywords(str, enum.Enum):
    ASSIGNMENT_DEPRECATED = r"^#+\s+Exercise `?([\w-]+)`?"
    ASSIGNMENT_START = r"^#+\s+Assignment `?([\w-]+)`?"
    SOLUTION = r"^###+\s+Solution"
    SKELETON = r"^###+\s+Skeleton"
    SOLUTION_END = r"^###+\s*$"
    HEADER = r"^#"
    IMAGE_TAG = r'<img[^>]*src="([^"]*)"'
    MD_IMAGE_TAG = r"!\[([^\]]*)\]\(([^)]*)\)"
    ASSIGNMENT_REPLACE = r"#Assignment-"


class Tags(str, enum.Enum):
    SOLUTION = "solution"
    ASSIGNMENT = "assignment"
    ASSIGNMENT_START = "assignment-start"
    SKELETON = "skeleton"
    NORMAL = "normal"


def cell_preview(cell):
    short = cell.source[0:100]
    text = re.sub(r"[\s\n]+", " ", short)
    return "'%s...'" % text[0:50]


def has_tag(cell, tag):
    ltag = tag.lower()
    if "tags" not in cell.metadata:
        return False
    for tag in cell.metadata.tags:
        if tag.lower() == ltag:
            return True
    return False


def set_tag(cell, tag):
    if has_tag(cell, tag):
        return
    if "tags" not in cell.metadata:
        cell.metadata.tags = []
    cell.metadata.tags.append(tag)


def to_snake_case(string):
    return string.replace(r"\s+", "_").lower()


def string_to_byte(input):
    units = {
        "B": 1,
        "KB": 1024,
        "MB": 1024**2,
        "GB": 1024**3,
        "TB": 1024**4,
        "KiB": 1024,
        "MiB": 1024**2,
        "GiB": 1024**3,
        "TiB": 1024**4,
    }
    match = re.match(r"(\d+(\.\d+)?)\s*(\w+)", input)
    if match:
        value, _, unit = match.groups()
        value = float(value)

        # Check if the unit is valid for conversion
        if unit in units:
            return int(value * units[unit])
        else:
            raise ValueError("Invalid unit")

    else:
        raise ValueError("Invalid format")



from nbconvert.preprocessors.base import Preprocessor

import os
import re
import requests
import urnc.preprocessor.util as util
from urnc.preprocessor.util import Keywords, Tags
import urnc.logger as log


def url_is_valid(url):
    try:
        response = requests.head(url, allow_redirects=True)
        is_valid = response.status_code == 200
        if not is_valid:
            log.warn(f"Got response code {response.status_code}")
        return is_valid
    except Exception:
        return False


def check_image(base_folder, src):
    max_size_str = "250 KiB"
    max_size = util.string_to_byte(max_size_str)
    if src.startswith("http"):
        log.warn(f"Remote image detected. {src}")
        if not url_is_valid(src):
            log.warn(f"Request to {src} failed.")
        return
    image_path = os.path.normpath(os.path.join(base_folder, src))
    if os.path.exists(image_path):
        try:
            image_size = os.path.getsize(image_path)
            if image_size > max_size:
                log.warn(f"The image {image_path} is larger than {max_size_str} KiB. This will increase the loading time of your notebook. Consider compressing the image.")
        except Exception:
            log.warn(f"Could not retrieve the size of image {image_path}.")
    else:
        log.warn(f"The image {image_path} does not exists.")


class Linter(Preprocessor):
    def preprocess(self, notebook, resources):
        path = resources["path"]

        base_folder = os.path.dirname(path)

        for cell in notebook.cells:
            matches = re.findall(Keywords.IMAGE_TAG, cell.source)
            for match in matches:
                log.dbg(f"Detected html image tag with src {match}")
                check_image(base_folder, match)

            matches = re.findall(Keywords.MD_IMAGE_TAG, cell.source)
            for match in matches:
                (alt, src) = match
                log.dbg(f"Detected markdown image tag with src {src}")
                check_image(base_folder, src)

        return notebook, resources
